#!/usr/bin/perl 
#####################################################
#
# SnS-Align.pl (Version 1.0) is protein alignment program using both structure and sequence
# Copyright (C) 2005 Ganiraju C.Manyam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
#####################################################
#
use strict;
use Tk;
use Tk::FileSelect;
use Tk::ROText;
use Tk::Button;
use Tk::BrowseEntry;
use Tk::Dialog;
use Tk::ProgressBar;
use File::Basename;
use Cwd;
#
my ($textpad1,$textpad2,$submit,$P_label,$GP_label,$GP_value,$GE_label,$GE_value,$m_list,$matrix_selected);
my ($file1,$file2,$file3,$tfile,$path,$mb,$geox,$geoy);
my $mw;my ($mw2,$canvas1,$canvas2);my ($alignagain,$savealign);
my ($GP,$GE,$null_label1,$null_label2,$null_label3);
my $th;
#
# The routine checkFasta is used to check whether
# a file is in a FASTA file or not
#
# It will also validate the protein Sequence
sub checkFasta
{
	my ($line,$fflag,$tfile,$aflag);
	$fflag=0;$tfile=$_[0];$aflag=1;
	open CHKIN,$_[0];
	$line=<CHKIN>;
	if($line !~ m/^(>)/)
	{
		$mb = $canvas1 -> messageBox(-message=>"File $tfile is not FASTA formatted",
               -type=>'ok',-title=>"SnS-Align");
				if($mb)
				{
					close CHKIN;
					if($tfile eq $file1)
					{
						$file1="";
						$textpad1->delete('1.0', 'end');
					}
					elsif($tfile eq $file2)
					{
						$file2="";
						$textpad2->delete('1.0', 'end');
					}
				}
				return -1;
	}
	close CHKIN;
	open CHKIN,$tfile;
	while($line=<CHKIN>)
	{
		chomp $line;
		if($line=~m/^(\s*)$/)
		{
			next;
		}
		elsif(($aflag==1) && ($line=~m/^(>)/))
		{
			$aflag=0;
			$fflag=1;
			next;
		}
		elsif(($aflag==0) && ($line=~m/^(>)/))
		{
			close CHKIN;
				$mb = $canvas1 -> messageBox(-message=>"In File $tfile, Protein sequence is missing or is not FASTA formatted",
		                -type=>'ok',-title=>"SnS-Align");
				if($mb)
				{
					if($tfile eq $file1)
					{
						$file1="";
						$textpad1->delete('1.0', 'end');
					}
					elsif($tfile eq $file2)
					{
						$file2="";
						$textpad2->delete('1.0', 'end');
					}
				}
				return -1;
		}
		elsif($fflag==0)
		{
				close CHKIN;
				$mb = $canvas1 -> messageBox(-message=>"File $tfile is not FASTA formatted",
		                -type=>'ok',-title=>"SnS-Align");
				if($mb)
				{
					if($tfile eq $file1)
					{
						$file1="";
						$textpad1->delete('1.0', 'end');
					}
					elsif($tfile eq $file2)
					{
						$file2="";
						$textpad2->delete('1.0', 'end');
					}
				}
				return -1;
		}
		if($fflag==1)
		{
			if($line=~m/^((\s*[A-Z]+\s*)+)$/i)
			{
				if($line=~m/(B|J|O|U|X|Z)/i)
				{
					close CHKIN;
					$mb = $mw -> messageBox(-message=>"Unknown amino acid \"$&\" in File $tfile",
		            					-type=>'ok',-title=>"SnS-Align");
				if($mb)
				{
					if($tfile eq $file1)
					{
						$file1="";
						$textpad1->delete('1.0', 'end');
					}
					elsif($tfile eq $file2)
					{
						$file2="";
						$textpad2->delete('1.0', 'end');
					}
				}
						return -1;
				}
				else
				{
					$aflag=1;
				}
			}
			else
			{
				close CHKIN;
				$mb = $mw -> messageBox(-message=>"Unknown protein sequence in File $tfile",
		            					-type=>'ok',-title=>"SnS-Align");
				if($mb)
				{
					if($tfile eq $file1)
					{
						$file1="";
						$textpad1->delete('1.0', 'end');
					}
					elsif($tfile eq $file2)
					{
						$file2="";
						$textpad2->delete('1.0', 'end');
					}
				}
						return -1;
			}
		}
	}
	close CHKIN;
}
#
# The routine parse_garnier is used to parse the
# secondary structure obtained from the garnier
# program of EMBOSS package
#
sub parse_garnier
{
	open GAIN,$_[0];
	open GOUT,">$_[1]";
	my ($line,$seq_name,@seqarray,@aarray,$i,$seq);
	while($line=<GAIN>)
	{
		chomp $line;
		if($line =~ m/^(\# Sequence: )(\S+)(\s+)(from)/)
		{
			$seq_name=$2;
			print GOUT ">$seq_name\n";
		}
		if($line =~ m/^(\s{6}(\w+))/)
		{
			$line=$2;
			@seqarray=split("",$line);
			next;
		}
		if($line =~ m/^(helix)/)
		{
			$line =substr($line,6);
			@aarray=split("",$line);
			for($i=0;$i<@aarray;$i++) {
				if($aarray[$i] eq "H")
				{
					$seqarray[$i] = "B";
				}
			}
			next;
		}
		if($line =~ m/^(sheet)/)
		{
			$line =substr($line,6);
			@aarray=split("",$line);
			for($i=0;$i<@aarray;$i++) {
				if($aarray[$i] eq "E")
				{
					$seqarray[$i] = "Z";
				}
			}
			$seq=join('',@seqarray);
			print GOUT $seq."\n";
			next;
		}
	}
	close GAIN;
	close GOUT;
}
#
# The routine SS_align calls the garnier and
# water programs EMBOSS package
#
# It will also validate the inputs to GUI
#
sub ss_align {
my ($process,$tfile2);
if(!($file1 && $file2))
{
	$mw -> messageBox(-message=>"Query or DB not found",
	            					-type=>'ok',-title=>"SnS-Align");
	return;
}
if(checkFasta($file1)== -1)
	{
		return;
	}
if(checkFasta($file2)== -1)
	{
		return;
	}
if(($GE !~ m/^(\d+|(\d+\.\d+))$/) || ($GP !~ m/^(\d+|(\d+\.\d+))$/))
	{
		$mw -> messageBox(-message=>"Gap Penalty should be an integer or floating point number",
	            					-type=>'ok',-title=>"SnS-Align");
		return;
	}

	$canvas1->delete( $submit );
	$submit->destroy();
	$process=$canvas1->Label(-text => 'Processing, Please Wait..',
							 -font =>"LucidaConsole 12 bold",
							);
	$process -> grid(-row=>10,-column=>3,-columnspan=>3);
	$canvas1->grid();
	$canvas1->update;

$file3=$path."temp/ssalign";
$tfile=$path."temp/garnier_temp00.out";
system("garnier -sequence $file1 -outfile $tfile");
print "for Query\n\n";
parse_garnier($tfile,$path."temp/parse_garnier001.out");
system("garnier -sequence $file2 -outfile $tfile");
print "for DB\n\n";
parse_garnier($tfile,$path."temp/parse_garnier002.out");
$matrix_selected=$path."data/".$matrix_selected.".ss";
$tfile=$path."temp/parse_garnier001.out";
$tfile2=$path."temp/parse_garnier002.out";
open CLEANOUT,">$file3";
close CLEANOUT;
system("water $tfile $tfile2 -outfile $file3 -gapopen $GP -gapextend $GE -datafile $matrix_selected");
recreate_canvas1();
}
#
# The routine load_file1 is used to take the 
# query protein in
#
sub load_file1 {
   $file1= $canvas1->getOpenFile();
   #$textpad1->text="";
   $textpad1->delete('1.0', 'end');
   if($file1) {
   $textpad1->insert(
						'insert',sprintf($file1)
					);
   }
}
#
# The routine load_file2 is used to take the 
# database (of proteins) in
#
sub load_file2 {
   $file2= $canvas1->getOpenFile();
   $textpad2->delete('1.0', 'end');
   if($file2) {
   $textpad2->insert(
						'insert',sprintf($file2)
					);
   }
}
#
# The function displays the options to 
# save upon calling the save alignment
#
sub save_op
{
		my ($D,$choice);
  $D = $mw->Dialog(
                 -title=>"SnS-Align",
                 -text  => "Save as ..",
                 -default_button => 'Text',
                 -buttons        => ['Text','Post Script','Cancel']
                       );
  $choice=$D->Show;
  if($choice eq "Text")
	{
	  save_text();
	}
 elsif($choice eq "Post Script")
	{
		save_ps();
	}
 elsif($choice eq "Cancel")
	{
		destroy $D;
	}
}
#
# The routine saves the canvas in the text version
# in the given output file
#
sub save_text
{
	my ($line,$align,@temp,@temp2,$tagline,$ids);my ($acount,$flag,$tcount);
	$acount=$flag=$tcount=0;$tagline="";$ids="";
	my @types = (["Text Files", '.txt'],
               ["All Files", "*"] );
  my $file = $mw->getSaveFile(-filetypes => \@types,
				  -initialfile => 'textfile',
				  -defaultextension => '.txt');
  if(!$file)
	{
	  return;
	}
  open SAVEOUT,">$file";
  print SAVEOUT "SnS-Align Version 1.0\n# - implies alpha-Helix , \$ - implies beta-Sheet\n\n";
	open IN, $file3;
	while($line=<IN>)
	{
		chomp $line;
		if(!$line)
		{
			next;
		}
		if($line=~m/^(\s{5,}(\.|\:|\|)+)/)
		{
			next;
		}
		if($line=~m/^(\#)/)
		{
			if($line=~m/^(\#\sScore\:\s)(\d+)/)
			{
				$flag=1;
				$tagline=$tagline.", Score=".$2.$ids;
				$ids="";
			}
			if($line=~m/^(\#\sIdentity\:\s+)\d+\/\d+\s+\(\s*(\d+\.\d)\%\)/)
			{
				$ids=$ids." (Identity=".$2."%";
			}
			if($line=~m/^(\#\sSimilarity\:\s+)\d+\/\d+\s+\(\s*(\d+\.\d)\%\)/)
			{
				$ids=$ids.", Similarity=".$2."%)";
			}
			if($line=~m/^(\#\s1:\s(.*))/)
			{
				$tagline=$2;
			}
			if($line=~m/^(\#\s2:\s(.*))/)
			{
				$tagline=$tagline." with ".$2;
			}
			next;
		}
		if($flag==1)
		{
			$flag=0;
			$acount++;
			$align="\nAlignment_".$acount.":\t".$tagline;
			$tagline="";
			print SAVEOUT $align."\n";
		}
		if($line=~m/^(.*)\s+(\d+)\s+((\w|\-)+)\s+(\d+)/)
		{
			$tcount++;
			$line=$1;
			while($line =~ m/(\s|\t)$/)
			{
				chop $line;
			}
			$align=$3;
			if(length($line) <= 7)
			{
				$line=$line."\t";
			}
			$line="\n$line"."\t(".$2."-".$5.")\t";
			if(length($2)+length($5)<=4)
			{
				$line=$line."\t";
			}
			print SAVEOUT $line;
			@temp2=@temp;
			@temp=split("",$align);
			foreach $line (@temp) {
			if($line eq "B")
				{
				print SAVEOUT "#";
				next;
				}
			if($line eq "Z")
				{
				print SAVEOUT "\$";
				next;
				}
			print SAVEOUT $line;
			}
			if($tcount==2)
			{
				$tcount=0;$line="";
				for($align=0;$align<@temp;$align++)
				{
					if($temp[$align] eq $temp2[$align])
					{
						if(($temp[$align] eq "B")|($temp[$align] eq "Z"))
						{
							$line=$line."*";
						}
						else
						{
							$line=$line."+";
						}
					}
					else
					{
						$line=$line." ";
					}
				}
				$line=(" "x32).$line;
				print SAVEOUT "\n$line\n";
			}
		}
	}
	close IN;
	close SAVEOUT;
}
#
# The routine saves the canvas into a postscript file
# by using the Tk function postscript
#
sub save_ps {
  my @types = (["Post Script Files", '.ps'],
               ["All Files", "*"] );
  my $file = $mw->getSaveFile(-filetypes => \@types,
				  -initialfile => 'imagefile',
				  -defaultextension => '.ps');
  $textpad2->postscript(-x=>0,-y=>0,
						-width => $textpad2->width,
                        -height => $th,
						-file =>$file);
 
}
#
# Create the initial canvas which takes
# input parameters
#
sub create_canvas1 {
	my (@matrices,$temp);
	$mw->destroy();
    $mw = MainWindow->new(-title=>"SnS-Align");
	$mw->resizable(0,0);
	$canvas1 = $mw->Scrolled('Canvas',
                       -scrollbars => 'd',
					   -width => 400,
					   -height => 700,
                       );
		$textpad1 = $canvas1->Scrolled(
				'ROText',
				-scrollbars => 'd',
				-width => 50,
				-height => 1,
				);
     $file1 = $canvas1->Button(-text => 'Query',-width => 10, -height=>1,-font =>"LucidaConsole 8 bold",-command =>  \&load_file1);
		 $textpad2 = $canvas1->Scrolled(
				'ROText',
				-scrollbars => 'd',
				-width => 50,
				-height => 1,
				);
	 $file2 = $canvas1->Button(-text => 'DB',-font =>"LucidaConsole 8 bold",-width => 10, -command =>  \&load_file2);
	 $submit=  $canvas1->Button(-text => 'SnS Align', -width=>20,-command =>  \&ss_align,-font =>"LucidaConsole 8 bold");

	$GP=10.0;$GE=0.5;
	$P_label= $canvas1->Label( -text => 'GAP Penalty',
	);
	$GP_label= $canvas1->Label( -text => 'Opening',
	);
	$GP_value= $canvas1->Entry(-width => 10,-textvariable => \$GP);
	
	$GE_label= $canvas1->Label( -text => 'Extension',
	);
	$GE_value= $canvas1->Entry(-width => 10,-textvariable => \$GE);

	$m_list= $canvas1->BrowseEntry(
		-label => "Scoring Matrix",
		-variable => \$matrix_selected,
			);
	@matrices=glob($path."data/*.ss");
	foreach (@matrices) {
		$temp=substr($_,rindex($_,"/")+1,rindex($_,".")-rindex($_,"/")-1);
		$m_list->insert('end', $temp);
	}
	$matrix_selected = "BLOSUM62";
	$null_label1= $canvas1->Label();
	$null_label2= $canvas1->Label();
	$null_label3= $canvas1->Label();
	$file1 -> grid(-row=>2,-column=>1,-columnspan=>2);
	$file2 -> grid(-row=>4,-column=>1,-columnspan=>2);
	$textpad1 -> grid(-row=>2,-column=>3,-columnspan=>4);
	$textpad2 -> grid(-row=>4,-column=>3,-columnspan=>4);
	$m_list->grid(-row=>6,-column=>5,-columnspan=>2);
	$null_label1->grid(-row=>5,-column=>1);
	$null_label2->grid(-row=>7,-column=>1);
	$null_label3->grid(-row=>9,-column=>1);
	$P_label->grid(-row=>8,-column=>1);
   	$GP_label->grid(-row=>8,-column=>3,-sticky=>'e');
	$GP_value->grid(-row=>8,-column=>4,-sticky=>'w');
	$GE_label->grid(-row=>8,-column=>5,-sticky=>'e');
	$GE_value->grid(-row=>8,-column=>6,-sticky=>'w');
	$submit -> grid(-row=>10,-column=>4,-columnspan=>2);
	$canvas1->grid();
	$geox=int(($mw->screenwidth-400)/2);
	$geoy=int(($mw->screenheight-$mw->reqheight)/2);
	$mw->geometry("+$geox+$geoy");
	$file1=$file2="";
}
#
# The routine creates the ouput canvas after parsing
# the temporary data file
#
sub recreate_canvas1
{
	my ($line,$align,@temp,@temp2,$tagline,$ids);my ($x,$y,$acount,$flag,$tcount);
	$y=20; $acount=$flag=$tcount=0;$tagline="";$ids="";
	$canvas2 = $mw->Scrolled('Canvas',
                       -scrollbars => 'd',
					   -scrollregion => [0, 0, 600,800],
					   -width => 600,
					   -height => 800,
					   -background => 'lightgreen',
					   -borderwidth => 20
                       );
	$textpad2 = $canvas2->Scrolled(
				'Canvas',
				-scrollregion => [0, 0, 700,800],
				-scrollbars => 'e',
				-background => 'white',
				-width => 700,
				-height =>400
				);
	open IN, $file3;
	$textpad2->createText(0, $y, -text=>"SnS-Align Version 1.0", -fill => 'red',-anchor => 'w', -font=>'CourierNew 10');
	$y=$y+25;
	$line=<IN>;
	if($line !~ m/^(\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#)$/)
	{
	$textpad2->createText(0, $y, -text=>"Error, occured in processing. Try again", -fill => 'red',-anchor => 'w', -font=>'CourierNew 10');

	}
	
	while($line=<IN>)
	{
		chomp $line;
		if(!$line)
		{
			next;
		}
		if($line=~m/^(\s{5,}(\.|\:|\|)+)/)
		{
			next;
		}
		if($line=~m/^(\#)/)
		{
			if($line=~m/^(\#\sScore\:\s)(\d+)/)
			{
				$flag=1;
				$tagline=$tagline.", Score=".$2.$ids;
				$ids="";
			}
			if($line=~m/^(\#\sIdentity\:\s+)\d+\/\d+\s+\(\s*(\d+\.\d)\%\)/)
			{
				$ids=$ids." (Identity=".$2."%";
			}
			if($line=~m/^(\#\sSimilarity\:\s+)\d+\/\d+\s+\(\s*(\d+\.\d)\%\)/)
			{
				$ids=$ids.", Similarity=".$2."%)";
			}
			if($line=~m/^(\#\s1:\s(.*))/)
			{
				$tagline=$2;
			}
			if($line=~m/^(\#\s2:\s(.*))/)
			{
				$tagline=$tagline." with ".$2;
			}
			next;
		}
		if($flag==1)
		{
			$flag=0;
			if($acount > 0)
			{
				$y=$y+40;
			}
			$acount++;
			$align="Alignment_".$acount.":\t".$tagline;
			if(length($tagline) >=88)
			{
				$tagline=substr($align,0,index($align,","));
				$textpad2->createText(0, $y, -text=>$tagline, -fill => 'magenta',-anchor => 'w', -font=>'CourierNew 10');
				$tagline=substr($align,index($align,",")+2);
				$y=$y+20;
				$textpad2->createText(110, $y, -text=>$tagline, -fill => 'magenta',-anchor => 'w', -font=>'CourierNew 10');
			}
			else
			{
				$textpad2->createText(0, $y, -text=>$align, -fill => 'magenta',-anchor => 'w', -font=>'CourierNew 10');
			}
			$tagline="";
		}
		$y=$y+20;
		if($line=~m/^(.*)\s+(\d+)\s+((\w|\-)+)\s+(\d+)/)
		{
			$tcount++;
			$line=$1;
			while($line =~ m/(\s|\t)$/)
			{
				chop $line;
			}
			$align=$3;
			if(length($line) <= 7)
			{
				$line=$line."\t";
			}
			$line=$line."\t(".$2."-".$5.")";
			if(length($2)+length($5)<=6)
			{
				$line=$line."\t";
			}
			$textpad2->createText(0, $y, -text=>$line, -fill => 'blue',-anchor => 'w', -font=>'LucidaConsole 10',-width=>199);
			@temp2=@temp;
			@temp=split("",$align);
			$x=200;
			foreach $line (@temp) {
			if($line eq "B")
				{
				$textpad2->createLine($x+2, $y-2, $x+5, $y+2,-fill=>'purple',-width=>3);
				$textpad2->createLine($x+2, $y+2, $x+5, $y-2,-fill=>'purple',-width=>3);
				$textpad2->createArc($x-2,$y-4, $x+2, $y,-start=>0,-extent=>90,-style=>"arc",-outline => 'purple',-width=>3);
				$textpad2->createArc($x-2,$y, $x+2, $y+4,-start=>0,-extent=>-90,-style=>"arc",-outline => 'purple',-width=>3);
				$textpad2->createArc($x+5,$y-4,$x+9, $y,-start=>180,-extent=>-90,-style=>"arc",-outline => 'purple',-width=>3);
				$textpad2->createArc($x+5,$y, $x+9, $y+4,-start=>180,-extent=>90,-style=>"arc",-outline => 'purple',-width=>3);
				$textpad2->createLine($x, $y+2, $x, $y+2,-fill=>'purple',-width=>3);
				$textpad2->createLine($x-2, $y-4, $x+1, $y-4,-fill=>'purple',-width=>3);
				$textpad2->createLine($x-2, $y+4, $x+1, $y+4,-fill=>'purple',-width=>3);
				$x=$x+9;
				next;
				}
			if($line eq "Z")
				{
				$textpad2->createLine($x, $y-2, $x+8, $y-2,-fill=>'lightblue',-width=>3);
				$textpad2->createLine($x, $y+2, $x+8, $y+2,-fill=>'lightblue',-width=>3);
				$x=$x+9;
				next;
				}
			$textpad2->createText($x, $y, -text=>$line, -fill => 'blue',-anchor => 'w', -font=>'LucidaConsole 8 bold');
			$x=$x+9;
			}
			$x=200;
			if($tcount==2)
			{
				$y=$y+20;
				$tcount=0;$line="";
				for($align=0;$align<@temp;$align++)
				{
					if($temp[$align] eq $temp2[$align])
					{
						if(($temp[$align] eq "B")|($temp[$align] eq "Z"))
						{
							$line=$line."*";
							$textpad2->createText($x, $y, -text=>"*", -fill => 'red',-anchor => 'w', -font=>'LucidaConsole 12');
							$x=$x+9;
						}
						else
						{
							$line=$line."+";
							$textpad2->createText($x, $y, -text=>"+", -fill => 'red',-anchor => 'w', -font=>'LucidaConsole 10');
							$x=$x+9;
						}
					}
					else
					{
						$line=$line." ";
						$x=$x+9;
					}
				}
			}
		}
	}
	$textpad2->configure(-scrollregion => [0, 0, 700,$y+20]);
	$canvas2->configure(-scrollregion => [0, 0, 700,$y+20]);
	$th=$y+20;
	close IN;
	$alignagain=  $canvas2->Button(-text => 'New Alignment',-width => 15, -command =>  \&create_canvas1);
	$savealign=  $canvas2->Button(-text => 'Save Alignment',-width => 15, -command =>  \&save_op);

	$textpad2-> grid(-row=>0,-column=>1,-columnspan=>3);
	$alignagain-> grid(-row=>2,-column=>1);
	$savealign-> grid(-row=>2,-column=>3);
	$canvas1->destroy();
	$canvas2->grid();
	$geox=int(($mw->screenwidth-720)/2);
	$geoy=int(($mw->screenheight-400)/2);
	$mw->geometry("+$geox+$geoy");
	$mw->update;

}
($tfile,$path) = fileparse($0, qr/\..*/); # Get the directory of the program
$mw = MainWindow->new;
create_canvas1(); 
MainLoop;
###################################################

